
/*
 * Copyright (C) Roman Arutyunyan
 */


#ifndef _NGX_RTMP_STREAMS_H_INCLUDED_
#define _NGX_RTMP_STREAMS_H_INCLUDED_


#define NGX_RTMP_MSID                   1

#define NGX_RTMP_CSID_AMF_INI           3
#define NGX_RTMP_CSID_AMF               5
#define NGX_RTMP_CSID_AUDIO             6
#define NGX_RTMP_CSID_VIDEO             7


#endif /* _NGX_RTMP_STREAMS_H_INCLUDED_ */
